/******************************************************************************/
/*                                                                            */
/* Broadcom AC1000 Linux Network Driver, Copyright (c) 2001 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/******************************************************************************/


char ac1000_driver[] = "ac1000";
char ac1000_version[] = "1.0.3";
char ac1000_date[] = "(12/27/01)";

#define AC1000UM
#include "ac1000mm.h"

#define TASKLET

/* A few user-configurable values. */

#define MAX_UNITS 16
/* Used to pass the full-duplex flag, etc. */
static int line_speed[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int full_duplex[MAX_UNITS] = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
static int rx_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int tx_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static int auto_flow_control[MAX_UNITS] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (2*HZ)

#if (LINUX_VERSION_CODE < 0x02030d)
#define pci_resource_start(dev, bar)	(dev->base_address[bar] & PCI_BASE_ADDRESS_MEM_MASK)
#elif (LINUX_VERSION_CODE < 0x02032b)
#define pci_resource_start(dev, bar)	(dev->resource[bar] & PCI_BASE_ADDRESS_MEM_MASK)
#endif

#if (LINUX_VERSION_CODE < 0x02032b)
#define dev_kfree_skb_irq(skb)  dev_kfree_skb(skb)
#define netif_wake_queue(dev)	clear_bit(0, &dev->tbusy); mark_bh(NET_BH)
#define netif_stop_queue(dev)	set_bit(0, &dev->tbusy)

static inline void netif_start_queue(struct net_device *dev)
{
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
}

#define netif_queue_stopped(dev)	dev->tbusy
#define netif_running(dev)		dev->start

static inline void tasklet_schedule(struct tasklet_struct *tasklet)
{
	queue_task(tasklet, &tq_immediate);
	mark_bh(IMMEDIATE_BH);
}

static inline void tasklet_init(struct tasklet_struct *tasklet,
				void (*func)(unsigned long),
				unsigned long data)
{
        tasklet->next = NULL;
        tasklet->sync = 0;
        tasklet->routine = (void (*)(void *))func;
        tasklet->data = (void *)data;
}

#define tasklet_disable(tasklet)
#define tasklet_kill(tasklet)

#endif

#if (LINUX_VERSION_CODE < 0x020300)
struct pci_device_id {
	unsigned int vendor, device;		/* Vendor and device ID or PCI_ANY_ID */
	unsigned int subvendor, subdevice;	/* Subsystem ID's or PCI_ANY_ID */
	unsigned int class, class_mask;		/* (class,subclass,prog-if) triplet */
	unsigned long driver_data;		/* Data private to the driver */
};

#define PCI_ANY_ID		0

#define pci_set_drvdata(pdev, dev)
#define pci_get_drvdata(pdev) 0

#define pci_enable_device(pdev) 0

#define __devinit		__init
#define __devinitdata		__initdata
#define __devexit

#define SET_MODULE_OWNER(dev)
#define MODULE_DEVICE_TABLE(pci, pci_tbl)

#endif


#if (LINUX_VERSION_CODE < 0x02032a)
static inline void *pci_alloc_consistent(struct pci_dev *pdev, size_t size,
					 dma_addr_t *dma_handle)
{
	void *virt_ptr;

	/* Maximum in slab.c */
	if (size > 131072)
		return 0;

	virt_ptr = kmalloc(size, GFP_KERNEL);
	*dma_handle = virt_to_bus(virt_ptr);
	return virt_ptr;
}
#define pci_free_consistent(cookie, size, ptr, dma_ptr)	kfree(ptr)
#define pci_map_single(cookie, address, size, dir)	virt_to_bus(address)
#define pci_unmap_single(cookie, address, size, dir)

#endif

#if (LINUX_VERSION_CODE < 0x020329)
#define pci_set_dma_mask(pdev, mask) (0)
#else
#if (LINUX_VERSION_CODE < 0x020403)
int
pci_set_dma_mask(struct pci_dev *dev, dma_addr_t mask)
{
    if(! pci_dma_supported(dev, mask))
        return -EIO;

    dev->dma_mask = mask;

    return 0;
}
#endif
#endif

#if (LINUX_VERSION_CODE < 0x020402)
#define pci_request_regions(pdev, name) (0)
#define pci_release_regions(pdev)
#endif

#define set_64bit_addr(paddr, low, high) \
	(paddr)->Low = low;		\
	(paddr)->High = high;

static inline void ac1000_set_addr(LM_PHYSICAL_ADDRESS *paddr, dma_addr_t addr)
{
	unsigned long baddr = (unsigned long) addr;
#if (BITS_PER_LONG == 64)
	set_64bit_addr(paddr, baddr & 0xffffffff, baddr >> 32);
#else
	set_64bit_addr(paddr, baddr, 0);
#endif
}

inline long
ac1000_lock(PUM_DEVICE_BLOCK pUmDevice)
{
	long flags;
	
	if (pUmDevice->do_global_lock) {
		spin_lock_irqsave(&pUmDevice->global_lock, flags);
		return flags;
	}
	return 0;
}

inline void
ac1000_unlock(PUM_DEVICE_BLOCK pUmDevice, long flags)
{
	if (pUmDevice->do_global_lock) {
		spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
	}
}

inline long
ac1000_tx_lock(PUM_DEVICE_BLOCK pUmDevice)
{
	long flags;
	
	if (pUmDevice->do_global_lock) {
		spin_lock_irqsave(&pUmDevice->global_lock, flags);
		return flags;
	}
	return 0;
}

inline void
ac1000_tx_unlock(PUM_DEVICE_BLOCK pUmDevice, long flags)
{
	if (pUmDevice->do_global_lock) {
		spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
	}
}

inline void
ac1000_intr_lock(PUM_DEVICE_BLOCK pUmDevice)
{
	if (pUmDevice->do_global_lock) {
		spin_lock(&pUmDevice->global_lock);
	}
}

inline void
ac1000_intr_unlock(PUM_DEVICE_BLOCK pUmDevice)
{
	if (pUmDevice->do_global_lock) {
		spin_unlock(&pUmDevice->global_lock);
	}
}

typedef struct _UM_PACKET {
	LM_PACKET lm_packet;
	struct sk_buff *skbuff;
	dma_addr_t map;
	size_t map_len;
	LM_FRAG_LIST frag_list;
} UM_PACKET, *PUM_PACKET;

int AC1000_Packet_Desc_Size = sizeof(UM_PACKET);

#if defined(MODULE)
MODULE_AUTHOR("Broadcom Corporation");
MODULE_DESCRIPTION("AC1000 Driver");
MODULE_PARM(debug, "i");
MODULE_PARM(line_speed, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(rx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(tx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(auto_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
#endif

#define RUN_AT(x) (jiffies + (x))

char kernel_version[] = UTS_RELEASE;

#define PCI_SUPPORT_VER2

#if ! defined(CAP_NET_ADMIN)
#define capable(CAP_XXX) (suser())
#endif

#define ac1000_debug debug
#ifdef AC1000_DEBUG
static int ac1000_debug = AC1000_DEBUG;
#else
static int ac1000_debug = 1;
#endif


#if DBG
#define STATIC
#else
#define STATIC static
#endif

STATIC int ac1000_open(struct net_device *dev);
STATIC void ac1000_timer(unsigned long data);
STATIC void ac1000_tx_timeout(struct net_device *dev);
STATIC int ac1000_start_xmit(struct sk_buff *skb, struct net_device *dev);
STATIC void ac1000_interrupt(int irq, void *dev_instance, struct pt_regs *regs);
STATIC void ac1000_tasklet(unsigned long data);
STATIC int ac1000_close(struct net_device *dev);
STATIC struct net_device_stats *ac1000_get_stats(struct net_device *dev);
STATIC int ac1000_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
STATIC void ac1000_set_rx_mode(struct net_device *dev);
STATIC int ac1000_set_mac_addr(struct net_device *dev, void *p);
STATIC int ac1000_reload_rx_buffers(PUM_DEVICE_BLOCK pUmDevice);
STATIC int ac1000_check_4G(PUM_DEVICE_BLOCK pUmDevice, PUM_PACKET pUmPacket);
STATIC int ac1000_freemem(struct net_device *dev);


/* A list of all installed ac1000 devices. */
static struct net_device *root_ac1000_dev = NULL;

typedef enum {
	AC91000A1 = 0,
	AC91000A1SX,
	AC91001A1,
	AC91002A1,
	AC91003A1
} board_t;


/* indexed by board_t, above */
static struct {
	char *name;
} board_info[] __devinitdata = {
	{ "Altima AC1000 1000Base-T" },
	{ "Altima AC1000 1000Base-SX" },
	{ "Altima AC1001 1000Base-T" },
	{ "Altima AC1002 1000Base-T" },
	{ "Altima AC1003 1000Base-T" },
	{ 0 },
	};

static struct pci_device_id ac1000_pci_tbl[] __devinitdata = {
	{0x173b, 0x03e8, 0x173b, 0x03e8, 0, 0, AC91000A1 },
	{0x173b, 0x03e8, 0x173b, 0x0003, 0, 0, AC91000A1SX },
	{0x173b, 0x03e9, 0x173b, 0x0001, 0, 0, AC91001A1 },
	{0x173b, 0x03ea, 0x173b, 0x0001, 0, 0, AC91002A1 },
	{0x173b, 0x03eb, 0x173b, 0x0002, 0, 0, AC91003A1 },
	{0,}
};

MODULE_DEVICE_TABLE(pci, ac1000_pci_tbl);

static int __devinit ac1000_init_board(struct pci_dev *pdev,
					struct net_device **dev_out,
					int board_idx)
{
	struct net_device *dev;
	PUM_DEVICE_BLOCK pUmDevice;
	PLM_DEVICE_BLOCK pDevice;
	int rc;

	*dev_out = NULL;

	/* dev zeroed in init_etherdev */
	dev = init_etherdev(NULL, sizeof(*pUmDevice));
	if (dev == NULL) {
		printk (KERN_ERR "%s: unable to alloc new ethernet\n",
			ac1000_driver);
		return -ENOMEM;
	}
	SET_MODULE_OWNER(dev);
	pUmDevice = (PUM_DEVICE_BLOCK) dev->priv;

	/* enable device (incl. PCI PM wakeup), and bus-mastering */
	rc = pci_enable_device (pdev);
	if (rc)
		goto err_out;

	rc = pci_request_regions(pdev, ac1000_driver);
	if (rc)
		goto err_out;

	pci_set_master(pdev);

	if (pci_set_dma_mask(pdev, ~(0UL)) != 0) {
		printk(KERN_ERR "System does not support DMA\n");
		pci_release_regions(pdev);
		goto err_out;
	}

	pUmDevice->dev = dev;
	pUmDevice->pdev = pdev;
	pUmDevice->mem_list_num = 0;
	pUmDevice->next_module = root_ac1000_dev;
	root_ac1000_dev = dev;
	pUmDevice->index = board_idx;

	spin_lock_init(&pUmDevice->global_lock);

	spin_lock_init(&pUmDevice->undi_lock);


	pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	if (AC1000_GetAdapterInfo(pDevice) != LM_STATUS_SUCCESS) {
		printk(KERN_ERR "Get Adapter info failed\n");
		rc = -ENODEV;
		goto err_out_unmap;
	}

	pUmDevice->do_global_lock = 0;
	if (T3_ASIC_REV(pUmDevice->lm_dev.ChipRevId) == T3_ASIC_REV_AC1000) {
		/* The 1000 chip works best without interleaved register */
		/* accesses on certain machines. */
		pUmDevice->do_global_lock = 1;
	}
	dev->mem_start = pci_resource_start(pdev, 0);
	dev->mem_end = dev->mem_start + pDevice->MemBaseSize;
	dev->irq = pDevice->Irq = pdev->irq;

	*dev_out = dev;
	return 0;

err_out_unmap:
	pci_release_regions(pdev);
	ac1000_freemem(dev);

err_out:
	unregister_netdev(dev);
	kfree (dev);
	return rc;
}

static int __devinit
ac1000_print_ver(void)
{
	printk(KERN_INFO "Altima Gigabit Ethernet Driver %s ",
		ac1000_driver);
	printk("ver. %s %s\n", ac1000_version, ac1000_date);
	return 0;
}

static int __devinit
ac1000_init_one(struct pci_dev *pdev,
				       const struct pci_device_id *ent)
{
	struct net_device *dev = NULL;
	PUM_DEVICE_BLOCK pUmDevice;
	PLM_DEVICE_BLOCK pDevice;
	int i;
	static int board_idx = -1;
	static int printed_version = 0;

	board_idx++;

	if (!printed_version) {
		ac1000_print_ver();
		printed_version = 1;
	}

	i = ac1000_init_board(pdev, &dev, board_idx);
	if (i < 0) {
		return i;
	}

	if (dev == NULL)
		return -ENOMEM;

	dev->open = ac1000_open;
	dev->hard_start_xmit = ac1000_start_xmit;
	dev->stop = ac1000_close;
	dev->get_stats = ac1000_get_stats;
	dev->set_multicast_list = ac1000_set_rx_mode;
	dev->do_ioctl = ac1000_ioctl;
	dev->set_mac_address = &ac1000_set_mac_addr;
#if (LINUX_VERSION_CODE >= 0x20400)
	dev->tx_timeout = ac1000_tx_timeout;
	dev->watchdog_timeo = TX_TIMEOUT;
#endif

	pUmDevice = (PUM_DEVICE_BLOCK) dev->priv;
	pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	dev->base_addr = pci_resource_start(pdev, 0);
	dev->irq = pdev->irq;

	pci_set_drvdata(pdev, dev);

	memcpy(dev->dev_addr, pDevice->NodeAddress, 6);
	pUmDevice->name = board_info[ent->driver_data].name,
	printk(KERN_INFO "%s: %s found at mem %lx, IRQ %d, ",
		dev->name, pUmDevice->name, dev->base_addr,
		dev->irq);
	printk("node addr ");
	for (i = 0; i < 6; i++) {
		printk("%2.2x", dev->dev_addr[i]);
	}
	printk("\n");

	printk(KERN_INFO "%s: ", dev->name);
	if ((pDevice->PhyId & PHY_ID_MASK) == PHY_BCM5400_PHY_ID)
		printk("Broadcom BCM5400 Copper ");
	else if ((pDevice->PhyId & PHY_ID_MASK) == PHY_BCM5401_PHY_ID)
		printk("Broadcom BCM5401 Copper ");
	else if ((pDevice->PhyId & PHY_ID_MASK) == PHY_BCM5411_PHY_ID)
		printk("Broadcom BCM5411 Copper ");
	else if ((pDevice->PhyId & PHY_ID_MASK) == PHY_AC1001_PHY_ID)
		printk("Altima AC1001 Integrated Copper ");
	else if (pDevice->EnableTbi)
		printk("Agilent HDMP-1636 SerDes ");
	else
		printk("Unknown ");
	printk("transceiver found\n");

#ifdef TASKLET
	tasklet_init(&pUmDevice->tasklet, ac1000_tasklet,
		(unsigned long) pUmDevice);
#endif
	return 0;

}


static void __devexit
ac1000_remove_one (struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata (pdev);
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;

	unregister_netdev(dev);

	if (pUmDevice->lm_dev.pMappedMemBase)
		iounmap(pUmDevice->lm_dev.pMappedMemBase);

	pci_release_regions(pdev);

	kfree(dev);

	pci_set_drvdata(pdev, NULL);

/*	pci_power_off(pdev, -1);*/

}

int __devinit
ac1000_probe(struct net_device *dev)
{
	int cards_found = 0;
	struct pci_dev *pdev = NULL;
	struct pci_device_id *pci_tbl;
	u16 ssvid, ssid;

	if ( ! pci_present())
		return -ENODEV;

	pci_tbl = ac1000_pci_tbl;
	while ((pdev = pci_find_class(PCI_CLASS_NETWORK_ETHERNET << 8, pdev))) {
		int idx;

		pci_read_config_word(pdev, PCI_SUBSYSTEM_VENDOR_ID, &ssvid);
		pci_read_config_word(pdev, PCI_SUBSYSTEM_ID, &ssid);
		for (idx = 0; pci_tbl[idx].vendor; idx++) {
			if ((pci_tbl[idx].vendor == PCI_ANY_ID ||
				pci_tbl[idx].vendor == pdev->vendor) &&
				(pci_tbl[idx].device == PCI_ANY_ID ||
				pci_tbl[idx].device == pdev->device) &&
				(pci_tbl[idx].subvendor == PCI_ANY_ID ||
				pci_tbl[idx].subvendor == ssvid) &&
				(pci_tbl[idx].subdevice == PCI_ANY_ID ||
				pci_tbl[idx].subdevice == ssid))
			{

				break;
			}
		}
		if (pci_tbl[idx].vendor == 0)
			continue;


		if (ac1000_init_one(pdev, &pci_tbl[idx]) == 0)
			cards_found++;
	}

	return cards_found ? 0 : -ENODEV;
}



STATIC int
ac1000_open(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	int index;

	index = pUmDevice->index;

	/* delay for 4 seconds */
	pUmDevice->delayed_link_ind = (4 * HZ) / pUmDevice->timer_interval;

	/* Sometimes we get spurious ints. after reset when link is down. */
	/* This field tells the isr to service the int. even if there is */
	/* no status block update. */
	if (pDevice->Bcm540xMode != BCM540X_MODE_LINK10) {
		pUmDevice->adapter_just_inited = (3 * HZ) /
			pUmDevice->timer_interval;
	}
	else {
		pUmDevice->adapter_just_inited = 0;
	}

	if (request_irq(dev->irq, &ac1000_interrupt, SA_SHIRQ, dev->name, dev)) {
		return -EAGAIN;
	}

	pUmDevice->opened = 1;

	if (AC1000_InitializeAdapter(pDevice) != LM_STATUS_SUCCESS) {
		free_irq(dev->irq, dev);
		ac1000_freemem(dev);
		return -EAGAIN;
	}

	if (memcmp(dev->dev_addr, pDevice->NodeAddress, 6)) {
		AC1000_SetMacAddress(pDevice, dev->dev_addr);
	}

	if (ac1000_debug > 1)
		printk(KERN_DEBUG "%s: ac1000_open() irq %d.\n", dev->name, dev->irq);

	QQ_InitQueue(&pUmDevice->rx_out_of_buf_q.Container,
        MAX_RX_PACKET_DESC_COUNT);
	netif_start_queue(dev);

#if (LINUX_VERSION_CODE < 0x020300)
	MOD_INC_USE_COUNT;
#endif

	init_timer(&pUmDevice->timer);
	pUmDevice->timer.expires = RUN_AT(pUmDevice->timer_interval);
	pUmDevice->timer.data = (unsigned long)dev;
	pUmDevice->timer.function = &ac1000_timer;
	add_timer(&pUmDevice->timer);

	AC1000_EnableInterrupt(pDevice);

	return 0;
}

STATIC void
ac1000_timer(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	long flags;

	if (!pUmDevice->opened)
		return;

	if (pUmDevice->delayed_link_ind > 0) {
		if (pUmDevice->delayed_link_ind == 1)
			AC1000_IndicateStatus(pDevice, pDevice->LinkStatus);
		else
			pUmDevice->delayed_link_ind--;
	}
	if (pUmDevice->adapter_just_inited > 0) {
		pUmDevice->adapter_just_inited--;
	}

	if (!pUmDevice->interrupt) {
		if (!pDevice->UseTaggedStatus) {
			flags = ac1000_lock(pUmDevice);
			if (pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) {
				/* This will generate an interrupt */
				REG_WR(pDevice, Grc.LocalCtrl,
					pDevice->GrcLocalCtrl |
					GRC_MISC_LOCAL_CTRL_SET_INT);
			}
			else {
				REG_WR(pDevice, HostCoalesce.Mode,
					HOST_COALESCE_ENABLE |
					HOST_COALESCE_NOW);
			}
			if (!(REG_RD(pDevice, DmaWrite.Mode) &
				DMA_WRITE_MODE_ENABLE)) {
				ac1000_tx_timeout(dev);
			}
			ac1000_unlock(pUmDevice, flags);
		}
#if (LINUX_VERSION_CODE < 0x02032b)
		if ((QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) !=
			pDevice->TxPacketDescCnt) &&
			((jiffies - dev->trans_start) > TX_TIMEOUT)) {

			printk(KERN_WARNING "%s: Tx hung\n", dev->name);
			ac1000_tx_timeout(dev);
		}
#endif
	}
#ifndef TASKLET
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container) >=
		pUmDevice->rx_buf_repl_panic_thresh) {
		/* Generate interrupt and let isr allocate buffers */
		REG_WR(pDevice, Grc.LocalCtrl, pDevice->GrcLocalCtrl |
			GRC_MISC_LOCAL_CTRL_SET_INT);
	}
#endif

	pUmDevice->timer.expires = RUN_AT(pUmDevice->timer_interval);
	add_timer(&pUmDevice->timer);
}

STATIC void
ac1000_tx_timeout(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	netif_stop_queue(dev);
	AC1000_DisableInterrupt(pDevice);
	AC1000_ResetAdapter(pDevice);	
	if (memcmp(dev->dev_addr, pDevice->NodeAddress, 6)) {
		AC1000_SetMacAddress(pDevice, dev->dev_addr);
	}
	AC1000_EnableInterrupt(pDevice);
	netif_wake_queue(dev);
}

STATIC int
ac1000_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	PLM_FRAG_LIST pfrag_list;
	long flags;
	unsigned int len;

	if ((pDevice->LinkStatus == LM_STATUS_LINK_DOWN) || !pDevice->InitDone)
	{
		dev_kfree_skb(skb);
		return 0;
	}

	
#if (LINUX_VERSION_CODE < 0x02032b)
	if (test_and_set_bit(0, &dev->tbusy)) {
		return 1;
	}
#endif

	if (pUmDevice->do_global_lock && pUmDevice->interrupt) {
		netif_stop_queue(dev);
		pUmDevice->tx_queued = 1;
		if (!pUmDevice->interrupt) {
			netif_wake_queue(dev);
			pUmDevice->tx_queued = 0;
		}
		return 1;
	}
	if ((atomic_read(&pDevice->SendBdLeft) == 0) ||
		((pPacket = (PLM_PACKET)
		QQ_PopHead(&pDevice->TxPacketFreeQ.Container)) == 0)) {

		netif_stop_queue(dev);
		pUmDevice->tx_full = 1;
		if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) &&
			atomic_read(&pDevice->SendBdLeft)) {
			netif_wake_queue(dev);
			pUmDevice->tx_full = 0;
		}
		return 1;
	}
	pUmPacket = (PUM_PACKET) pPacket;
	pUmPacket->skbuff = skb;
	pfrag_list = &pUmPacket->frag_list;
	pPacket->u.Tx.pFraglist = pfrag_list;
	pPacket->Flags = 0;
	len = skb->len;

	pfrag_list->Fragments[0].FragSize = len;
	pUmPacket->map = pci_map_single(pUmDevice->pdev,
		skb->data, len, PCI_DMA_TODEVICE);
	pUmPacket->map_len = len;
	ac1000_set_addr(&pfrag_list->Fragments[0].FragBuf,
		pUmPacket->map);

	pPacket->u.Tx.FragCount = 1;

	/* Work around 4GB dma problem */
	if (T3_ASIC_REV(pDevice->ChipRevId) == T3_ASIC_REV_AC1000) {
		if (ac1000_check_4G(pUmDevice, pUmPacket) == 0) {
			QQ_PushHead(&pDevice->TxPacketFreeQ.Container, pPacket);
#if (LINUX_VERSION_CODE < 0x02032b)
			netif_wake_queue(dev);
#endif
			return 0;
		}
		skb = pUmPacket->skbuff;
	}

	flags = ac1000_tx_lock(pUmDevice);
	AC1000_SendPacket(pDevice, pPacket);
	ac1000_tx_unlock(pUmDevice, flags);
	if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) == 0) {
		netif_stop_queue(dev);
		pUmDevice->tx_full = 1;
	}
#if (LINUX_VERSION_CODE < 0x02032b)
	else {
		netif_wake_queue(dev);
	}
#endif
	dev->trans_start = jiffies;
	return 0;
}

STATIC
int
ac1000_check_4G(PUM_DEVICE_BLOCK pUmDevice, PUM_PACKET pUmPacket)
{
	u32 base;
	int i;
	PLM_FRAG_LIST pfrag_list = &pUmPacket->frag_list;
	PLM_FRAG pfrag;
	struct sk_buff *skb, *nskb;
	int do_copy = 0;

	/* Work around 4GB dma problem */
	for (i = 0; i < pUmPacket->lm_packet.u.Tx.FragCount; i++) {
		pfrag = &pfrag_list->Fragments[i];

		if (((base = pfrag->FragBuf.Low) > 0xffffdcc0) &&
			(pfrag->FragBuf.High == 0) &&
			((base + 8 + pfrag->FragSize) < base)) {

			do_copy = 1;
			break;
		}
	}
	if (do_copy) {
		skb = pUmPacket->skbuff;
		if ((nskb = dev_alloc_skb(skb->len))) {
			memcpy(nskb->data, skb->data, pUmPacket->map_len);
			pci_unmap_single(pUmDevice->pdev,
				pUmPacket->map,
				pUmPacket->map_len, PCI_DMA_TODEVICE);
			nskb->len = skb->len;
			pUmPacket->map =
				pci_map_single(pUmDevice->pdev,
					nskb->data,
					nskb->len, PCI_DMA_TODEVICE);
			ac1000_set_addr(&pfrag_list->Fragments[0].FragBuf,
				pUmPacket->map);
			pfrag_list->FragCount =
				pUmPacket->lm_packet.u.Tx.FragCount = 1;
		}
		dev_kfree_skb(skb);
		pUmPacket->skbuff = nskb;
		if (nskb == 0) {
			return 0;
		}
	}
	return 1;
}


STATIC void
ac1000_interrupt(int irq, void *dev_instance, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *)dev_instance;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	LM_UINT32 oldtag, newtag;
	int repl_buf_count;

	if (!pDevice->InitDone)
		return;

	ac1000_intr_lock(pUmDevice);
	if (test_and_set_bit(0, (void*)&pUmDevice->interrupt)) {
		printk(KERN_ERR "%s: Duplicate entry of the interrupt handler by "
			   "processor %d.\n",
			   dev->name, hard_smp_processor_id());
		ac1000_intr_unlock(pUmDevice);
		return;
	}

	if (pDevice->UseTaggedStatus) {
		if ((pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) ||
			pUmDevice->adapter_just_inited) {
			REG_WR(pDevice, Mailbox.Interrupt[0].Low, 1);
			oldtag = pDevice->pStatusBlkVirt->StatusTag;

			while (1) {
   				pDevice->pStatusBlkVirt->Status &=
					~STATUS_BLOCK_UPDATED;

				AC1000_ServiceInterrupts(pDevice);
				newtag = pDevice->pStatusBlkVirt->StatusTag;
				if (newtag == oldtag) {
					REG_WR(pDevice,
						Mailbox.Interrupt[0].Low,
						newtag << 24);
					break;
				}
				oldtag = newtag;
			}
		}
	}
	else if ((pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) ||
			pUmDevice->adapter_just_inited) {
		do {
			REG_WR(pDevice, Mailbox.Interrupt[0].Low, 1);
   			pDevice->pStatusBlkVirt->Status &= ~STATUS_BLOCK_UPDATED;
			AC1000_ServiceInterrupts(pDevice);
			REG_WR(pDevice, Mailbox.Interrupt[0].Low, 0);
			REG_RD(pDevice, Mailbox.Interrupt[0].Low);
		}
		while (pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED);
	}
#ifdef TASKLET
	repl_buf_count = QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container);
	if (repl_buf_count >= pUmDevice->rx_buf_repl_thresh) {
		if ((repl_buf_count >= pUmDevice->rx_buf_repl_panic_thresh) &&
			(!test_and_set_bit(0, &pUmDevice->tasklet_busy))) {
			ac1000_reload_rx_buffers(pUmDevice);
			clear_bit(0, (void*)&pUmDevice->tasklet_busy);
		}
		else if (!pUmDevice->tasklet_pending) {
			pUmDevice->tasklet_pending = 1;
			tasklet_schedule(&pUmDevice->tasklet);
		}
	}
#else
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container)) {
		ac1000_reload_rx_buffers(pUmDevice);
	}

	if (QQ_GetEntryCnt(&pDevice->RxPacketFreeQ.Container)) {
		AC1000_QueueRxPackets(pDevice);
	}
#endif

	clear_bit(0, (void*)&pUmDevice->interrupt);
	ac1000_intr_unlock(pUmDevice);
	return;
}


STATIC void
ac1000_tasklet(unsigned long data)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)data;

	/* RH 7.2 Beta 3 tasklets are reentrant */
	if (test_and_set_bit(0, &pUmDevice->tasklet_busy)) {
		pUmDevice->tasklet_pending = 0;
		return;
	}

	pUmDevice->tasklet_pending = 0;
	ac1000_reload_rx_buffers(pUmDevice);
	clear_bit(0, &pUmDevice->tasklet_busy);
}

STATIC int
ac1000_close(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

#if (LINUX_VERSION_CODE < 0x02032b)
	dev->start = 0;
#endif
	netif_stop_queue(dev);
	pUmDevice->opened = 0;

	if (ac1000_debug > 1)
		printk(KERN_DEBUG "%s: Shutting down Tigon3\n",
			   dev->name);

	AC1000_DisableInterrupt(pDevice);
#ifdef TASKLET
	tasklet_disable(&pUmDevice->tasklet);
	tasklet_kill(&pUmDevice->tasklet);
#endif
	AC1000_Halt(pDevice);
	pDevice->InitDone = 0;
	del_timer(&pUmDevice->timer);

	free_irq(dev->irq, dev);
#if (LINUX_VERSION_CODE < 0x020300)
	MOD_DEC_USE_COUNT;
#endif
	ac1000_freemem(dev);

	return 0;
}

STATIC int
ac1000_freemem(struct net_device *dev)
{
	int i;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;

	for (i = 0; i < pUmDevice->mem_list_num; i++) {
		if (pUmDevice->dma_list[i] == 0) {
			kfree(pUmDevice->mem_list[i]);
		}
		else {
			pci_free_consistent(pUmDevice->pdev,
				(size_t) pUmDevice->mem_size_list[i],
				pUmDevice->mem_list[i],
				pUmDevice->dma_list[i]);
		}
	}
	pUmDevice->mem_list_num = 0;
	return 0;
}

LM_UINT32
ac1000_crc_count(PUM_DEVICE_BLOCK pUmDevice)
{
	PLM_DEVICE_BLOCK pDevice = &pUmDevice->lm_dev;
	LM_UINT32 Value32;
	PT3_STATS_BLOCK pStats = (PT3_STATS_BLOCK) pDevice->pStatsBlkVirt;
	unsigned long flags;

#if INCLUDE_TBI_SUPPORT
	if(pDevice->EnableTbi)
		return (pStats->dot3StatsFCSErrors.Low);
#endif
	if (T3_ASIC_REV(pDevice->ChipRevId) == T3_ASIC_REV_AC1000 ||
		T3_ASIC_REV(pDevice->ChipRevId) == T3_ASIC_REV_AC1001) {

		if (pUmDevice->adapter_just_inited)
			return 0;
		flags = ac1000_lock(pUmDevice);
		AC1000_ReadPhy(pDevice, 0x1e, &Value32);
		AC1000_WritePhy(pDevice, 0x1e, Value32 | 0x8000);
		AC1000_ReadPhy(pDevice, 0x14, &Value32);
		ac1000_unlock(pUmDevice, flags);
		pDevice->PhyCrcCount += Value32;
		return pDevice->PhyCrcCount;
	}
	else {
		return (pStats->dot3StatsFCSErrors.Low);
	}
}

STATIC struct net_device_stats *
ac1000_get_stats(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	PT3_STATS_BLOCK pStats = (PT3_STATS_BLOCK) pDevice->pStatsBlkVirt;
	struct net_device_stats *p_netstats = &pUmDevice->stats;

	if (pStats == 0)
		return p_netstats;

	/* Get stats from LM */
	p_netstats->rx_packets = pStats->ifHCInUcastPkts.Low +
			pStats->ifHCInMulticastPkts.Low +
			pStats->ifHCInBroadcastPkts.Low;
	p_netstats->tx_packets = pStats->COSIfHCOutPkts[0].Low;
	p_netstats->rx_bytes = pStats->ifHCInOctets.Low;
	p_netstats->tx_bytes = pStats->ifHCOutOctets.Low;
	p_netstats->rx_errors = pStats->ifInErrors.Low;
	p_netstats->tx_errors = pStats->dot3StatsInternalMacTransmitErrors.Low +
			pStats->dot3StatsCarrierSenseErrors.Low +
			pStats->ifOutDiscards.Low +
			pStats->ifOutErrors.Low;
	p_netstats->multicast = pStats->ifHCInMulticastPkts.Low;
	p_netstats->collisions = pStats->etherStatsCollisions.Low;
	p_netstats->rx_length_errors = pStats->dot3StatsFramesTooLong.Low +
			pStats->etherStatsUndersizePkts.Low;
	p_netstats->rx_over_errors = pStats->nicNoMoreRxBDs.Low;
	p_netstats->rx_crc_errors = ac1000_crc_count(pUmDevice);
	p_netstats->rx_frame_errors = pStats->dot3StatsAlignmentErrors.Low;
	p_netstats->tx_aborted_errors = pStats->ifOutDiscards.Low;
	p_netstats->tx_carrier_errors = pStats->dot3StatsCarrierSenseErrors.Low;

	return p_netstats;
}

#ifdef SIOCETHTOOL
static int netdev_ethtool_ioctl(struct net_device *dev, void *useraddr)
{
	struct ethtool_cmd ethcmd;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
		
	if (copy_from_user(&ethcmd, useraddr, sizeof(ethcmd)))
		return -EFAULT;

        switch (ethcmd.cmd) {
#ifdef ETHTOOL_GDRVINFO
        case ETHTOOL_GDRVINFO: {
		struct ethtool_drvinfo info = {ETHTOOL_GDRVINFO};

		strcpy(info.driver,  ac1000_driver);
		strcpy(info.version, ac1000_version);
		strcpy(info.bus_info, pUmDevice->pdev->slot_name);
		if (copy_to_user(useraddr, &info, sizeof(info)))
			return -EFAULT;
		return 0;
	}
#endif
        case ETHTOOL_GSET: {
		ethcmd.supported =
			(SUPPORTED_10baseT_Half | SUPPORTED_10baseT_Full |
			 SUPPORTED_100baseT_Half | SUPPORTED_100baseT_Full |
			 SUPPORTED_1000baseT_Half | SUPPORTED_1000baseT_Full |
			 SUPPORTED_Autoneg);
		if (pDevice->EnableTbi) {
			ethcmd.supported |= SUPPORTED_FIBRE;
			ethcmd.port = PORT_FIBRE;
		}
		else {
			ethcmd.supported |= SUPPORTED_TP;
			ethcmd.port = PORT_TP;
		}

		ethcmd.transceiver = XCVR_INTERNAL;
		ethcmd.phy_address = 0;

		if (pUmDevice->line_speed == 1000)
			ethcmd.speed = SPEED_1000;
		else if (pUmDevice->line_speed == 100)
			ethcmd.speed = SPEED_100;
		else if (pUmDevice->line_speed == 10)
			ethcmd.speed = SPEED_10;
		else
			ethcmd.speed = 0;

		if (pDevice->DuplexMode == LM_DUPLEX_MODE_FULL)
			ethcmd.duplex = DUPLEX_FULL;
		else
			ethcmd.duplex = DUPLEX_HALF;

		if (pDevice->DisableAutoNeg == FALSE)
			ethcmd.autoneg = AUTONEG_ENABLE;
		else
			ethcmd.autoneg = AUTONEG_DISABLE;

		ethcmd.maxtxpkt = pDevice->TxPacketDescCnt;
		ethcmd.maxrxpkt = pDevice->RxPacketDescCnt;

		if(copy_to_user(useraddr, &ethcmd, sizeof(ethcmd)))
			return -EFAULT;
		return 0;
	}
	case ETHTOOL_SSET: {
		if(!capable(CAP_NET_ADMIN))
			return -EPERM;
		if (ethcmd.autoneg == AUTONEG_ENABLE) {
			if (pDevice->EnableTbi) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_AUTO;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_AUTO;
			}
			pDevice->DisableAutoNeg = FALSE;
		}
		else if (ethcmd.speed == SPEED_1000) {
			if (!pDevice->EnableTbi)
				return -EINVAL;
			pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_UTP_1000MBPS_FULL_DUPLEX;
			pDevice->DisableAutoNeg = TRUE;
		}
		else if (ethcmd.speed == SPEED_100) {
			if (ethcmd.duplex == DUPLEX_FULL) {
				pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS;
			}
			pDevice->DisableAutoNeg = TRUE;
		}
		else if (ethcmd.speed == SPEED_10) {
			if (ethcmd.duplex == DUPLEX_FULL) {
				pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS;
			}
			pDevice->DisableAutoNeg = TRUE;
		}
		else {
			return -EINVAL;
		}
		AC1000_SetupPhy(pDevice);
		return 0;
	}
#ifdef ETHTOOL_GWOL
	case ETHTOOL_GWOL: {
		struct ethtool_wolinfo wol = {ETHTOOL_GWOL};

		wol.supported = 0;
		wol.wolopts = 0;
		if (copy_to_user(useraddr, &wol, sizeof(wol)))
			return -EFAULT;
		return 0;
	}
	case ETHTOOL_SWOL: {
		return -EINVAL;
        }
#endif
	}
	
	return -EOPNOTSUPP;
}
#endif
/* Provide ioctl() calls to examine the MII xcvr state. */
STATIC int ac1000_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	u16 *data = (u16 *)&rq->ifr_data;
	u32 value;
	unsigned long flags;

	switch(cmd) {
	case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
		data[0] = pDevice->PhyAddr;
	case SIOCDEVPRIVATE+1:		/* Read the specified MII register. */
		flags = ac1000_lock(pUmDevice);
		AC1000_ReadPhy(pDevice, data[1] & 0x1f, (LM_UINT32 *) &value);
		ac1000_unlock(pUmDevice, flags);
		data[3] = value & 0xffff;
		return 0;
	case SIOCDEVPRIVATE+2:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
		flags = ac1000_lock(pUmDevice);
		AC1000_WritePhy(pDevice, data[1] & 0x1f, data[2]);
		ac1000_unlock(pUmDevice, flags);
		return 0;
#ifdef SIOCETHTOOL
	case SIOCETHTOOL:
		return netdev_ethtool_ioctl(dev, (void *) rq->ifr_data);
#endif
	default:
		return -EOPNOTSUPP;
	}
	return -EOPNOTSUPP;
}

STATIC void ac1000_set_rx_mode(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	int i;
	struct dev_mc_list *mclist;

	AC1000_MulticastClear(pDevice);
	for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next) {
		AC1000_MulticastAdd(pDevice, (PLM_UINT8) &mclist->dmi_addr);
	}
	if (dev->flags & IFF_ALLMULTI) {
		if (!(pDevice->ReceiveMask & LM_ACCEPT_ALL_MULTICAST)) {
			AC1000_SetReceiveMask(pDevice,
				pDevice->ReceiveMask | LM_ACCEPT_ALL_MULTICAST);
		}
	}
	else if (pDevice->ReceiveMask & LM_ACCEPT_ALL_MULTICAST) {
		AC1000_SetReceiveMask(pDevice,
			pDevice->ReceiveMask & ~LM_ACCEPT_ALL_MULTICAST);
	}
	if (dev->flags & IFF_PROMISC) {
		if (!(pDevice->RxMode & RX_MODE_PROMISCUOUS_MODE)) {
			AC1000_SetReceiveMask(pDevice,
				pDevice->ReceiveMask | LM_PROMISCUOUS_MODE);
		}
	}
	else if (pDevice->RxMode & RX_MODE_PROMISCUOUS_MODE) {
		AC1000_SetReceiveMask(pDevice,
			pDevice->ReceiveMask & ~LM_PROMISCUOUS_MODE);
	}
}

/*
 * Set the hardware MAC address.
 */
STATIC int ac1000_set_mac_addr(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) dev->priv;

	if (netif_running(dev))
		return -EBUSY;
	memcpy(dev->dev_addr, addr->sa_data,dev->addr_len);
	AC1000_SetMacAddress(pDevice, dev->dev_addr);
	return 0;
}


#if (LINUX_VERSION_CODE < 0x020300)
#ifdef MODULE
int init_module(void)
{
	return ac1000_probe(NULL);
}

void cleanup_module(void)
{
	struct net_device *next_dev;
	PUM_DEVICE_BLOCK pUmDevice;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_ac1000_dev) {
		pUmDevice = (PUM_DEVICE_BLOCK)root_ac1000_dev->priv;
		next_dev = pUmDevice->next_module;
		unregister_netdev(root_ac1000_dev);
		if (pUmDevice->lm_dev.pMappedMemBase)
			iounmap(pUmDevice->lm_dev.pMappedMemBase);
		kfree(root_ac1000_dev);
		root_ac1000_dev = next_dev;
	}
}

#endif  /* MODULE */
#else	/* LINUX_VERSION_CODE < 0x020300 */

#if (LINUX_VERSION_CODE >= 0x020406)
static int ac1000_suspend (struct pci_dev *pdev, u32 state)
#else
static void ac1000_suspend (struct pci_dev *pdev)
#endif
{
	struct net_device *dev = (struct net_device *) pci_get_drvdata(pdev);
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	if (!netif_running(dev))
#if (LINUX_VERSION_CODE >= 0x020406)
		return 0;
#else
		return;
#endif

	AC1000_DisableInterrupt(pDevice);
	netif_device_detach (dev);

	/* Disable interrupts, stop Tx and Rx. */
	AC1000_Halt(pDevice);

/*	pci_power_off(pdev, -1);*/
#if (LINUX_VERSION_CODE >= 0x020406)
	return 0;
#endif
}


#if (LINUX_VERSION_CODE >= 0x020406)
static int ac1000_resume(struct pci_dev *pdev)
#else
static void ac1000_resume(struct pci_dev *pdev)
#endif
{
	struct net_device *dev = (struct net_device *) pci_get_drvdata(pdev);
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) dev->priv;

	if (!netif_running(dev))
#if (LINUX_VERSION_CODE >= 0x020406)
		return 0;
#else
		return;
#endif
/*	pci_power_on(pdev);*/
	netif_device_attach(dev);
	AC1000_InitializeAdapter(pDevice);
	if (memcmp(dev->dev_addr, pDevice->NodeAddress, 6)) {
		AC1000_SetMacAddress(pDevice, dev->dev_addr);
	}
	AC1000_EnableInterrupt(pDevice);
#if (LINUX_VERSION_CODE >= 0x020406)
	return 0;
#endif
}


static struct pci_driver ac1000_pci_driver = {
	name:		ac1000_driver,
	id_table:	ac1000_pci_tbl,
	probe:		ac1000_init_one,
	remove:		ac1000_remove_one,
	suspend:	ac1000_suspend,
	resume:		ac1000_resume,
};


static int __init ac1000_init_module (void)
{
	return pci_module_init(&ac1000_pci_driver);
}


static void __exit ac1000_cleanup_module (void)
{
	pci_unregister_driver(&ac1000_pci_driver);
}


module_init(ac1000_init_module);
module_exit(ac1000_cleanup_module);
#endif

/*
 * Middle Module
 *
 */


LM_STATUS
AC1000_ReadConfig16(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT16 *pValue16)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_read_config_word(pUmDevice->pdev, Offset, (u16 *) pValue16);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_ReadConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 *pValue32)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_read_config_dword(pUmDevice->pdev, Offset, (u32 *) pValue32);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_WriteConfig16(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT16 Value16)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_write_config_word(pUmDevice->pdev, Offset, Value16);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_WriteConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 Value32)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_write_config_dword(pUmDevice->pdev, Offset, Value32);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_AllocateSharedMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt, PLM_PHYSICAL_ADDRESS pMemoryBlockPhy,
	LM_BOOL Cached)
{
	PLM_VOID pvirt;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	dma_addr_t mapping;

	pvirt = pci_alloc_consistent(pUmDevice->pdev, BlockSize,
					       &mapping);
	if (!pvirt)
		return LM_STATUS_FAILURE;
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = mapping;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = BlockSize;
	memset(pvirt, 0, BlockSize);
	*pMemoryBlockVirt = (PLM_VOID) pvirt;
	ac1000_set_addr(pMemoryBlockPhy, mapping);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_AllocateMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt)
{
	PLM_VOID pvirt;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;


	/* Maximum in slab.c */
	if (BlockSize > 131072)
		return LM_STATUS_FAILURE;

	pvirt = kmalloc(BlockSize, GFP_KERNEL);
	if (!pvirt)
		return LM_STATUS_FAILURE;
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = 0;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = BlockSize;
	memset(pvirt, 0, BlockSize);
	*pMemoryBlockVirt = pvirt;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_MapMemBase(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;

	pDevice->pMappedMemBase = ioremap_nocache(
		pci_resource_start(pUmDevice->pdev, 0), pDevice->MemBaseSize);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_InitializeUmPackets(PLM_DEVICE_BLOCK pDevice)
{
	int i;
	struct sk_buff *skb;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	PUM_PACKET pUmPacket;
	PLM_PACKET pPacket;

	for (i = 0; i < pDevice->RxPacketDescCnt; i++) {
		pPacket = QQ_PopHead(&pDevice->RxPacketFreeQ.Container);
		pUmPacket = (PUM_PACKET) pPacket;
		if (pPacket == 0) {
			printk(KERN_DEBUG "Bad RxPacketFreeQ\n");
		}
		skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize + 2);
		if (skb == 0) {
			pUmPacket->skbuff = 0;
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
			continue;
		}
		pUmPacket->skbuff = skb;
		skb->dev = pUmDevice->dev;
		skb_reserve(skb, 2);
		pUmPacket->map = pci_map_single(pUmDevice->pdev,
					skb->tail,
					pPacket->u.Rx.RxBufferSize,
					PCI_DMA_FROMDEVICE);
		pPacket->u.Rx.pRxBufferVirt = skb->tail;
		ac1000_set_addr(&pPacket->u.Rx.RxBufferPhy,
			pUmPacket->map);
		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}
	pUmDevice->rx_buf_repl_thresh = pDevice->RxPacketDescCnt / 8;
	pUmDevice->rx_buf_repl_panic_thresh = pDevice->RxPacketDescCnt / 2;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_GetConfig(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	int index = pUmDevice->index;

	if (line_speed[index] == 0) {
		pDevice->RequestedMediaType =
				LM_REQUESTED_MEDIA_TYPE_AUTO;
		pDevice->DisableAutoNeg = FALSE;
	}
	else {
		pDevice->DisableAutoNeg = TRUE;
		if (line_speed[index] == 1000) {
			if (pDevice->EnableTbi) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_FIBER_1000MBPS_FULL_DUPLEX;
			}
			else if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_1000MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_1000MBPS;
			}
			if (!pDevice->EnableTbi)
				pDevice->DisableAutoNeg = FALSE;
		}
		else if (line_speed[index] == 100) {
			if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS;
			}
		}
		else if (line_speed[index] == 10) {
			if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS;
			}
		}
		else {
			pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_AUTO;
			pDevice->DisableAutoNeg = FALSE;
		}

	}
	pDevice->FlowControlCap = 0;
	if (rx_flow_control[index] != 0) {
		pDevice->FlowControlCap |= LM_FLOW_CONTROL_RECEIVE_PAUSE;
	}
	if (tx_flow_control[index] != 0) {
		pDevice->FlowControlCap |= LM_FLOW_CONTROL_TRANSMIT_PAUSE;
	}
	if ((auto_flow_control[index] != 0) &&
		(pDevice->DisableAutoNeg == FALSE)) {

		pDevice->FlowControlCap |= LM_FLOW_CONTROL_AUTO_PAUSE;
		if ((tx_flow_control[index] == 0) && (rx_flow_control[index] == 0)) {
			pDevice->FlowControlCap |=
					LM_FLOW_CONTROL_TRANSMIT_PAUSE |
					LM_FLOW_CONTROL_RECEIVE_PAUSE;
		}
	}

	if (T3_ASIC_REV(pDevice->ChipRevId) == T3_ASIC_REV_AC1000) {
		pDevice->UseTaggedStatus = TRUE;
		pUmDevice->timer_interval = HZ;
	}
	else {
		pUmDevice->timer_interval = HZ/10;
	}

	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_IndicateRxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->RxPacketReceivedQ.Container);
		if (pPacket == 0)
			break;
		pUmPacket = (PUM_PACKET) pPacket;
		pci_unmap_single(pUmDevice->pdev,
				pUmPacket->map,
				pPacket->u.Rx.RxBufferSize,
				PCI_DMA_FROMDEVICE);
		if (pPacket->PacketStatus != LM_STATUS_SUCCESS) {

			/* reuse skb */
#ifdef TASKLET
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
#else
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
#endif
			continue;
		}
		skb = pUmPacket->skbuff;
		skb_put(skb, pPacket->PacketSize);
		skb->pkt_type = 0;
		skb->protocol = eth_type_trans(skb, skb->dev);
		skb->ip_summed = CHECKSUM_NONE;
		if ((pPacket->Flags & RCV_BD_FLAG_TCP_UDP_CHKSUM_FIELD) &&
			(pDevice->TaskToOffload &
				LM_TASK_OFFLOAD_RX_TCP_CHECKSUM) &&
			(pPacket->u.Rx.TcpUdpChecksum == 0xffff)) {

			skb->ip_summed = CHECKSUM_UNNECESSARY;

		}
		netif_rx(skb);

#ifdef TASKLET
		pUmPacket->skbuff = 0;
		QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
#else
		skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize + 2);
		if (skb == 0) {
			pUmPacket->skbuff = 0;
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
		}
		else {
			pUmPacket->skbuff = skb;
			skb->dev = pUmDevice->dev;
			skb_reserve(skb, 2);
			pUmPacket->map = pci_map_single(
						pUmDevice->pdev,
						skb->tail,
						pPacket->u.Rx.RxBufferSize,
						PCI_DMA_FROMDEVICE);
			pPacket->u.Rx.pRxBufferVirt = skb->tail;
			ac1000_set_addr(&pPacket->u.Rx.RxBufferPhy,
				pUmPacket->map[0].dma_map);
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
		}
#endif
	}
	return LM_STATUS_SUCCESS;
}

/* Returns 1 if not all buffers are allocated */
STATIC int
ac1000_reload_rx_buffers(PUM_DEVICE_BLOCK pUmDevice)
{
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	struct sk_buff *skb;
	int queue_rx = 0;
	int ret = 0;

	while ((pUmPacket = (PUM_PACKET)
		QQ_PopHead(&pUmDevice->rx_out_of_buf_q.Container)) != 0) {
		pPacket = (PLM_PACKET) pUmPacket;
		if (pUmPacket->skbuff) {
			/* reuse an old skb */
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			queue_rx = 1;
			continue;
		}
		if ((skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize + 2)) == 0) {
			QQ_PushHead(&pUmDevice->rx_out_of_buf_q.Container,
				pPacket);
			ret = 1;
			break;
		}
		pUmPacket->skbuff = skb;
		skb->dev = pUmDevice->dev;
		skb_reserve(skb, 2);
		pPacket->u.Rx.pRxBufferVirt = skb->tail;
		pUmPacket->map = pci_map_single(pUmDevice->pdev,
					skb->tail,
					pPacket->u.Rx.RxBufferSize,
					PCI_DMA_FROMDEVICE);
		ac1000_set_addr(&pPacket->u.Rx.RxBufferPhy,
			pUmPacket->map);
		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
		queue_rx = 1;
	}
	if (queue_rx) {
		AC1000_QueueRxPackets(pDevice);
	}
	return ret;
}

LM_STATUS
AC1000_IndicateTxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->TxPacketXmittedQ.Container);
		if (pPacket == 0)
			break;
		pUmPacket = (PUM_PACKET) pPacket;
		skb = pUmPacket->skbuff;
		pci_unmap_single(pUmDevice->pdev, pUmPacket->map,
				pUmPacket->map_len, PCI_DMA_TODEVICE);
		dev_kfree_skb_irq(skb);
		pUmPacket->skbuff = 0;
		QQ_PushTail(&pDevice->TxPacketFreeQ.Container, pPacket);
	}
	if (pUmDevice->tx_full) {
		if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) >=
			(pDevice->TxPacketDescCnt >> 1)) {

			pUmDevice->tx_full = 0;
			netif_wake_queue(pUmDevice->dev);
		}
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_IndicateStatus(PLM_DEVICE_BLOCK pDevice, LM_STATUS Status)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	struct net_device *dev = pUmDevice->dev;
	LM_FLOW_CONTROL flow_control;

	if (!pUmDevice->opened)
		return LM_STATUS_SUCCESS;

	if (pUmDevice->delayed_link_ind > 0) {
		pUmDevice->delayed_link_ind = 0;
		printk(KERN_ERR "%s: %s ", ac1000_driver, dev->name);
		if (Status == LM_STATUS_LINK_DOWN) {
			printk("NIC Link is DOWN\n");
		}
		else if (Status == LM_STATUS_LINK_ACTIVE) {
			printk("NIC Link is UP, ");
		}
	}
	else {
		printk(KERN_ERR "%s: %s ", ac1000_driver, dev->name);
		if (Status == LM_STATUS_LINK_DOWN) {
			pUmDevice->line_speed = 0;
			printk("NIC Link is Down\n");
		}
		else if (Status == LM_STATUS_LINK_ACTIVE) {
			printk("NIC Link is Up, ");
		}
	}

	if (Status == LM_STATUS_LINK_ACTIVE) {
		if (pDevice->LineSpeed == LM_LINE_SPEED_1000MBPS)
			pUmDevice->line_speed = 1000;
		else if (pDevice->LineSpeed == LM_LINE_SPEED_100MBPS)
			pUmDevice->line_speed = 100;
		else if (pDevice->LineSpeed == LM_LINE_SPEED_10MBPS)
			pUmDevice->line_speed = 10;

		printk("%d Mbps ", pUmDevice->line_speed);

		if (pDevice->DuplexMode == LM_DUPLEX_MODE_FULL)
			printk("full duplex");
		else
			printk("half duplex");

		flow_control = pDevice->FlowControl &
			(LM_FLOW_CONTROL_RECEIVE_PAUSE |
			LM_FLOW_CONTROL_TRANSMIT_PAUSE);
		if (flow_control) {
			if (flow_control & LM_FLOW_CONTROL_RECEIVE_PAUSE) {
				printk(", receive ");
				if (flow_control & LM_FLOW_CONTROL_TRANSMIT_PAUSE)
					printk("& transmit ");
			}
			else {
				printk(", transmit ");
			}
			printk("flow control ON");
		}
		printk("\n");
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
AC1000_FreeRxBuffer(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;

	if (pPacket == 0)
		return LM_STATUS_SUCCESS;
	pUmPacket = (PUM_PACKET) pPacket;
	if ((skb = pUmPacket->skbuff))
		dev_kfree_skb(skb);
	pUmPacket->skbuff = 0;
	return LM_STATUS_SUCCESS;
}

unsigned long
AC1000_AnGetCurrentTime_us(PAN_STATE_INFO pAnInfo)
{
	struct timeval tv;

	do_gettimeofday(&tv);
	return ((tv.tv_sec * 1000000) + tv.tv_usec);
}

#if (LINUX_VERSION_CODE >= 0x02040A)
MODULE_LICENSE("GPL");
#endif

